/*
 *   libaudiotag - A media file tag-reader library
 *   Copyright (C) 2003, 2004  Pipian, Audun Hove
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "id3v2.h"
#include "endian.h"
#include "../fmt.h"
#include "../unicode.h"
#define BUFFER_SIZE 4096

typedef struct
{
	unsigned char *check;
	int count;
} resync_t;

typedef struct
{
	int unsync, extended, size;
	char version[2];
} id3header_t;

static resync_t *checkunsync(char *syncCheck, int size)
{
	int i, j;
	resync_t *sync;
	
	sync = malloc(sizeof(resync_t));

	sync->check = syncCheck;
	sync->count = 0;
	
	if(size == 0)
		size = strlen(sync->check);

	for(i = 0; i < strlen(sync->check); i++)
	{
		if(sync->check[i] == 0xFF && sync->check[i + 1] == 0x00)
		{
			for(j = i + 1; j < size - 1; j++)
				syncCheck[j] = syncCheck[j + 1];
			sync->check[j] = '\0';
			sync->count++;
		}
	}

	return sync;
}

static void unsync(char *data, char *bp)
{
	resync_t *unsynced;
	char *syncFix = NULL;
	int i;
	
	unsynced = checkunsync(data, 0);
	while(unsynced->count > 0)
	{
		if(syncFix != NULL)
			syncFix = realloc(syncFix, unsynced->count);
		else
			syncFix = malloc(unsynced->count);
		memcpy(syncFix, bp, unsynced->count);
		bp += unsynced->count;
		for(i = 0; i < unsynced->count; i++)
			data[4 - unsynced->count + i] = syncFix[i];
		free(unsynced);
		unsynced = checkunsync(data, 0);
	}
	free(unsynced);
	free(syncFix);
}

/*
 * Header:
 *
 * identifier: 3 bytes "ID3" ("3DI" if footer)
 * version: 2 bytes
 * flags: 1 byte
 * tag size: 4 bytes
 */
static id3header_t *read_header(FILE *fp)
{
	id3header_t *id3_data = malloc(sizeof(id3header_t));
	char id3_flags, cToInt[4];
	int bottom = 0;
	
	memset(id3_data, 0, sizeof(id3header_t));
	
	fread(cToInt, 1, 3, fp);
	if(strncmp(cToInt, "3DI", 3) == 0)
		bottom = 1;
	fread(id3_data->version, 1, 2, fp);
	fread(&id3_flags, 1, 1, fp);
	if((id3_flags & 0x80) == 0x80)
		id3_data->unsync = 1;
	if((id3_flags & 0x40) == 0x40 && id3_data->version[0] > 0x02)
		id3_data->extended = 1;
	fread(cToInt, 1, 4, fp);
	id3_data->size = synchsafe2int(cToInt);
	if(bottom == 1)
		fseek(fp, -10 - id3_data->size, SEEK_CUR);
		
	if(id3_data->version[0] == 0x04)
		pdebug("Version: ID3v2.4");
	else if(id3_data->version[0] == 0x03)
		pdebug("Version: ID3v2.3");
	else if(id3_data->version[0] == 0x02)
		pdebug("Version: ID3v2.2");
	else
		return NULL;
		
	return id3_data;
}

static int id3_lookupframe(char *tag, int tagver)
{
	int i;

	switch (tagver) {
	case ID3v22:
		for (i = 0; id3v22_lookup[i].frameid; i++)
			if (!strcmp(tag, id3v22_lookup[i].frameid))
				return id3v22_lookup[i].code;
		return -1;
		break;
	case ID3v23:
		for (i = 0; id3v23_lookup[i].frameid; i++)
			if (!strcmp(tag, id3v23_lookup[i].frameid))
				return id3v23_lookup[i].code;
		return -1;
		break;
	case ID3v24:
		for (i = 0; id3v23_lookup[i].frameid; i++)
			if (!strcmp(tag, id3v24_lookup[i].frameid))
				return id3v24_lookup[i].code;
		return -1;
		break;
	}
	return -1;
}

static framedata_t *parseFrame(char *bp, char *end, id3header_t *id3_data)
{
	static unsigned char frameid[5];
	unsigned char frameflags[2] = "", cToInt[5];
	int framesize, frameidcode;
	framedata_t *framedata;

	/* TODO: Unsync, decompress, decrypt, grouping, data length */
	switch (id3_data->version[0]) {
	case 2:
		if (end - bp < 6)
			return NULL;
		frameid[3] = 0;
		memcpy(frameid, bp, 3);
		bp += 3;
		frameidcode = id3_lookupframe(frameid, ID3v22);
		memcpy(cToInt, bp, 3);
		cToInt[3] = 0;
		if(id3_data->unsync)
			unsync(cToInt, bp);
		framesize = be24int(cToInt);
		bp += 3;
		break;
	case 3:
		if (end - bp < 10)
			return NULL;
		frameid[4] = 0;
		memcpy(frameid, bp, 4);
		bp += 4;
		frameidcode = id3_lookupframe(frameid, ID3v23);
		memcpy(cToInt, bp, 4);
		if(id3_data->unsync)
			unsync(cToInt, bp);
		framesize = be2int(cToInt);
		bp += 4;

		memcpy(frameflags, bp, 2);
		bp += 2;
		break;
	case 4:
		if (end - bp < 10)
			return NULL;
		frameid[4] = 0;
		memcpy(frameid, bp, 4);
		bp += 4;
		frameidcode = id3_lookupframe(frameid, ID3v24);
		memcpy(cToInt, bp, 4);
		framesize = synchsafe2int(cToInt);
		bp += 4;

		memcpy(frameflags, bp, 2);
		bp += 2;
		break;
	default:
		/* Should not be reached */
		return NULL;
	}

	/* printf("found id:%s, size:%-8d\n", frameid, framesize); */
	if (framesize > end - bp)
		return NULL;
	framedata = malloc(sizeof(framedata_t));
	memset(framedata, 0, sizeof(framedata_t));
	framedata->frameid = frameidcode;
	if(id3_data->unsync)
		frameflags[1] |= 0x02;
	framedata->flags = malloc(2);
	memcpy(framedata->flags, frameflags, 2);
	framedata->len = framesize;
	framedata->data = malloc(framesize);
	if(id3_data->version[0] == 0x04)
	{
		/*
		 * 4 bytes extra for compression or original size
		 * 1 byte extra for encyption
		 * 1 byte extra for grouping
		 */
		if((frameflags[1] & 0x08) == 0x08 ||
			(frameflags[1] & 0x01) == 0x01)
			bp += 4;
		if((frameflags[1] & 0x04) == 0x04)
			bp++;
		if((frameflags[1] & 0x40) == 0x40)
			bp++;
	}
	else if(id3_data->version[0] == 0x03)
	{
		/*
		 * 4 bytes extra for compression or original size
		 * 1 byte extra for encyption
		 * 1 byte extra for grouping
		 */
		if((frameflags[1] & 0x80) == 0x80)
		{
			char tmp[4];
			
			memcpy(tmp, bp, 4);
			bp += 4;
			if((frameflags[1] & 0x02) == 0x02)
				unsync(tmp, bp);
		}
		if((frameflags[1] & 0x40) == 0x40)
			bp++;
		if((frameflags[1] & 0x20) == 0x20)
			bp++;
	}
	memcpy(framedata->data, bp, framesize);
	bp += framesize;
	return framedata;
}

static id3v2_t *readFrames(char *bp, char *end, id3header_t *id3_data)
{
	id3v2_t *id3v2 = malloc(sizeof(id3v2_t));
	
	memset(id3v2, 0, sizeof(id3v2_t));

	while(bp - id3_data->size < end)
	{
		framedata_t *framedata = NULL;
		
		framedata = parseFrame(bp, end, id3_data);
		
		if(framedata == NULL)
			break;
		
		id3v2->items = realloc(id3v2->items, (id3v2->numitems + 1) * 
			sizeof(framedata_t *));
		id3v2->items[id3v2->numitems++] = framedata;
	}
	id3v2->version = id3_data->version[0];
	
	return id3v2;
}

unsigned char *ID3v2_parseText(framedata_t *frame)
{
	unsigned char *data = NULL, *utf = NULL, *ptr;
	char encoding;
	
	ptr = frame->data;
	
	encoding = *(ptr++);
	data = realloc(data, frame->len);
	memset(data, '\0', frame->len);
	memcpy(data, ptr, frame->len - 1);
	if((frame->flags[1] & 0x02) == 0x02)
	{
		resync_t *unsync = checkunsync(data, 0);
		free(unsync);
	}
	if(encoding == 0x00)
	{
		if(utf != NULL)
			free(utf);
		iso88591_to_utf8(data, &utf);
	}
	else if(encoding == 0x01)
	{
		if(utf != NULL)
			free(utf);
		utf16bom_to_utf8(data, frame->len - 1, &utf);
	}
	else if(encoding == 0x02)
	{
		if(utf != NULL)
			free(utf);
		utf16be_to_utf8(data, frame->len - 1, &utf);
	}
	else if(encoding == 0x03)
	{
		utf = realloc(utf, strlen(data) + 1);
		strcpy(utf, data);
	}
	
	return utf;
}

unsigned char *ID3v2_getData(framedata_t *frame)
{
	unsigned char *data = NULL, *ptr;
	
	ptr = frame->data;
	
	if(data != NULL)
		data = realloc(data, frame->len + 1);
	else
		data = malloc(frame->len + 1);
	memset(data, '\0', frame->len + 1);
	memcpy(data, ptr, frame->len);
	if((frame->flags[1] & 0x02) == 0x02)
	{
		resync_t *unsync = checkunsync(data, frame->len);
		free(unsync);
	}
	
	return data;
}

int findID3v2(FILE *fp)
{
	unsigned char tag_buffer[BUFFER_SIZE], *bp = tag_buffer;
	int pos, search = -1, i, status = 0, charsRead;
	
	charsRead = fread(tag_buffer, 1, 10, fp);
	pos = 0;
	bp = tag_buffer;
	while(status == 0)
	{
		if(search == -1)
		{
			if((strncmp(bp, "ID3", 3) == 0 ||
				strncmp(bp, "3DI", 3) == 0)) status = 1;
			fseek(fp, 3, SEEK_END);
			charsRead = fread(tag_buffer, 1, 3, fp);
			search = -2;
		}
		else
		{
			if(search == -2)
			{
				bp = tag_buffer;
				pos = ftell(fp);
				if((strncmp(bp, "ID3", 3) == 0 ||
					strncmp(bp, "3DI", 3) == 0)) status = 1;
				search = 1;
			}
			if(status != 1)
			{
				pos = ftell(fp) - BUFFER_SIZE;
				fseek(fp, pos, SEEK_SET);
				charsRead = fread(tag_buffer, 1, BUFFER_SIZE,
						fp);

				bp = tag_buffer;
				for(i = 0; i < charsRead - 3 && status == 0;
					i++)
				{
					bp++;
					if((strncmp(bp, "ID3", 3) == 0 ||
						strncmp(bp, "3DI", 3) == 0))
							status = 1;
				}
				if(status == 1) pos += bp - tag_buffer;
				pos -= BUFFER_SIZE - 9;
				if((pos < -BUFFER_SIZE + 9 || ferror(fp)) &&
					status != 1)
						status = -1;
			}
		}
		/*
		 * An ID3v2 tag can be detected with the following pattern:
		 *
		 * $49 44 33 yy yy xx zz zz zz zz
		 *
		 * Where yy is less than $FF, xx is the 'flags' byte and zz is 
		 * less than $80.
		 */
		if(status == 1 && *(bp + 3) < 0xFF && *(bp + 4) < 0xFF &&
			*(bp + 6) < 0x80 && *(bp + 7) < 0x80 &&
			*(bp + 8) < 0x80 && *(bp + 9) < 0x80)
				status = 1;
		else if(status != -1) status = 0;
		if(search == 0) search = -1;
	}
	if(status > 0)
		return -1;
	else
		return pos;
}

id3v2_t *readID3v2(char *filename)
{
	FILE *fp;
	id3v2_t *id3v2 = NULL;
	int pos;
	
	fp = fopen(filename, "rb");
	
	fseek(fp, 0, SEEK_SET);

	pdebug("Searching for tag...");
	pos = findID3v2(fp);
	if(pos > -1)
	{
		id3header_t *id3_data;
		id3v2_t *id3v2;
		char *tag_buffer, *bp;
		
		/* Found the tag. */
		fseek(fp, pos, SEEK_SET);
		pdebug("Found ID3v2 tag...");
		/* Read the header */
		id3_data = read_header(fp);
		if(id3_data == NULL)
		{
			pdebug("Or not.");
			fclose(fp);
			return NULL;
		}
		/* Read tag into buffer */
		tag_buffer = malloc(id3_data->size);
		fread(tag_buffer, 1, id3_data->size, fp);
		bp = tag_buffer;
		/* Skip extended header */
		if(id3_data->extended)
		{
			char cToInt[4];
			
			memcpy(cToInt, bp, 4);
			bp += 4;
			if(id3_data->version[0] == 0x03 &&
				id3_data->unsync == 1)
					unsync(cToInt, bp);
			if(id3_data->version[0] > 0x03)
				bp += synchsafe2int(cToInt);
			else
				bp += be2int(cToInt);
		}
		/* Read frames into id3v2_t */
		id3v2 = readFrames(bp, tag_buffer, id3_data);
	}
	
	fclose(fp);
	
	return id3v2;
}

void freeID3v2(id3v2_t *id3v2)
{
	int i;
	
	for(i = 0; i < id3v2->numitems; i++)
	{
		framedata_t *frame;
		
		frame = id3v2->items[i];
		free(frame->flags);
		free(frame->data);
		free(frame);
	}
	free(id3v2->items);
	free(id3v2);
}
